
;--- mini debugger
;--- test of MS debug engine API
;--- uses Win32Inc

	.386
	.model flat,stdcall
	option casemap:none

	.nolist
	.nocref
WIN32_LEAN_AND_MEAN equ 1
	include windows.inc
	include dbgeng.inc
	include stdio.inc
	include stdlib.inc
	.list
	.cref

	includelib kernel32.lib
	includelib msvcrt.lib
	includelib dbgeng.lib
	includelib uuid.lib

CStr macro text:REQ
local xxx
	.const
xxx	db text,0
	.code
	exitm <offset xxx>
	endm

dprintf macro fstr, items:VARARG
ifdef _DEBUG
	pushad
	ifnb <items>
		invoke printf, CStr(<fstr>), items
	else
		invoke printf, CStr(<fstr>)
	endif
	popad
endif
	endm

CDebugEventCallbacks struct
	IDebugEventCallbacks <>
cnt	dword ?
CDebugEventCallbacks ends

CDebugInputCallbacks struct
	IDebugInputCallbacks <>
cnt	dword ?
CDebugInputCallbacks ends

CDebugOutputCallbacks struct
	IDebugOutputCallbacks <>
cnt	dword ?
CDebugOutputCallbacks ends

	.data

pDebugClient  PDEBUG_CLIENT  NULL
pDebugControl PDEBUG_CONTROL NULL
pDebugSymbols PDEBUG_SYMBOLS NULL

;pDebugEventCallbacks  PDEBUG_EVENT_CALLBACKS  NULL
;pDebugInputCallbacks  PDEBUG_INPUT_CALLBACKS  NULL
;pDebugOutputCallbacks PDEBUG_OUTPUT_CALLBACKS NULL

opbuf DEBUG_CREATE_PROCESS_OPTIONS < DEBUG_PROCESS, 0, 0, 0 >
pszFile dd NULL
bDump db 0
bStateChanged db 0

	.const

DebugEventCallbacksVtbl label IDebugEventCallbacksVtbl
	dd QueryInterface@DebugEventCallbacks
	dd AddRef@DebugEventCallbacks
	dd Release@DebugEventCallbacks
	dd GetInterestMask@DebugEventCallbacks
	dd Breakpoint@DebugEventCallbacks
	dd Exception@DebugEventCallbacks
	dd CreateThread@DebugEventCallbacks
	dd ExitThread@DebugEventCallbacks
	dd CreateProcess@DebugEventCallbacks
	dd ExitProcess@DebugEventCallbacks
	dd LoadModule@DebugEventCallbacks
	dd UnloadModule@DebugEventCallbacks
	dd SystemError@DebugEventCallbacks
	dd SessionStatus@DebugEventCallbacks
	dd ChangeDebuggeeState@DebugEventCallbacks
	dd ChangeEngineState@DebugEventCallbacks
	dd ChangeSymbolState@DebugEventCallbacks
	
DebugInputCallbacksVtbl label IDebugInputCallbacksVtbl
	dd QueryInterface@DebugInputCallbacks
	dd AddRef@DebugInputCallbacks
	dd Release@DebugInputCallbacks
	dd StartInput@DebugInputCallbacks
	dd EndInput@DebugInputCallbacks

DebugOutputCallbacksVtbl label IDebugOutputCallbacksVtbl
	dd QueryInterface@DebugOutputCallbacks
	dd AddRef@DebugOutputCallbacks
	dd Release@DebugOutputCallbacks
	dd Output@DebugOutputCallbacks

szUsage label byte
	db "CDBAsm v0.77, Public Domain, japheth 2009.",10
	db "Mini-Debugger based on MS Debug Engine API.",10
	db "Needs a somewhat recent version of DbgEng.dll,",10
	db "which one might find in 'Debugging Tools for Windows'.",10
	db "usage: CDBAsm [ options ] filename",10
	db "options:",10
	db "  -z: assume file is a dump file",10
	db 0

	.code

GetArguments proc uses ebx esi argc:dword, argv:ptr ptr

	mov ebx, argc
	mov esi, argv
	add esi, 4
	dec ebx
	.while ( ebx )
		lodsd
		.if ( byte ptr [eax] == '-' || byte ptr [eax] == '/')
			inc eax
			mov cx,[eax]
			or cl,20h
			.if ( cx == 'z' )
				mov bDump,1
			.else
				dec eax
				invoke printf, CStr(<"unknown option %s",10>), eax
				xor eax, eax
				ret
			.endif
		.elseif ( pszFile == NULL )
			mov pszFile, eax
		.else
			invoke printf, CStr(<"more than one file argument given",10>)
			xor eax, eax
			ret
		.endif
		dec ebx
	.endw
	mov eax,1
	ret
	align 4
GetArguments endp

main proc c uses ebx argc:dword, argv:ptr ptr

local dwRead:dword
local buffer[128]:byte
local szCurrDir[260]:byte

	invoke GetArguments, argc, argv
	and eax, eax
	jz exit_
	.if ( pszFile == NULL )
		invoke printf, CStr("%s"), addr szUsage
		jmp exit_
	.endif

;--- create objects IDebugClient5, IDebugControl, IDebugSymbols

	invoke DebugCreate, addr IID_IDebugClient5, addr pDebugClient
	.if ( eax != S_OK )
		invoke printf, CStr(<"unable to create IDebugClient5 object. DbgEng.dll too old?",10>)
		jmp exit_
	.endif
	invoke vf( pDebugClient, IDebugClient, QueryInterface), addr IID_IDebugControl, addr pDebugControl
	.if ( eax != S_OK )
		invoke printf, CStr(<"IDebugClient:QueryInterface failed [%X]",10>), eax
		jmp exit_
	.endif
	invoke vf( pDebugClient, IDebugClient, QueryInterface), addr IID_IDebugSymbols, addr pDebugSymbols
	.if ( eax != S_OK )
		invoke printf, CStr(<"IDebugClient:QueryInterface failed [%X]",10>), eax
		jmp exit_
	.endif

;--- create and set callback objects Event, Input, Output

	invoke malloc, sizeof CDebugEventCallbacks
	.if ( !eax )
		invoke printf, CStr(<"out of memory",10>)
		jmp exit_
	.endif
	mov [eax].CDebugEventCallbacks.lpVtbl, offset DebugEventCallbacksVtbl
	mov [eax].CDebugEventCallbacks.cnt, 0
	invoke vf( pDebugClient, IDebugClient, SetEventCallbacks ), eax
	.if ( eax != S_OK )
		invoke printf, CStr(<"IDebugClient:SetEventCallbacks failed [%X]",10>), eax
	.endif

	invoke malloc, sizeof CDebugInputCallbacks
	.if ( !eax )
		invoke printf, CStr(<"out of memory",10>)
		jmp exit_
	.endif
	mov [eax].CDebugInputCallbacks.lpVtbl, offset DebugInputCallbacksVtbl
	mov [eax].CDebugInputCallbacks.cnt, 0
	invoke vf( pDebugClient, IDebugClient, SetInputCallbacks ), eax
	.if ( eax != S_OK )
		invoke printf, CStr(<"IDebugClient:SetInputCallbacks failed [%X]",10>), eax
	.endif

	invoke malloc, sizeof CDebugOutputCallbacks
	.if ( !eax )
		invoke printf, CStr(<"out of memory",10>)
		jmp exit_
	.endif
	mov [eax].CDebugOutputCallbacks.lpVtbl, offset DebugOutputCallbacksVtbl
	mov [eax].CDebugOutputCallbacks.cnt, 0
	invoke vf( pDebugClient, IDebugClient, SetOutputCallbacks ), eax
	.if ( eax != S_OK )
		invoke printf, CStr(<"IDebugClient:SetOutputCallbacks failed [%X]",10>), eax
	.endif
if 0
	invoke vf( pDebugClient, IDebugClient, SetOutputMask ), DEBUG_OUTPUT_NORMAL or DEBUG_OUTPUT_ERROR or DEBUG_OUTPUT_WARNING or DEBUG_OUTPUT_VERBOSE or DEBUG_OUTPUT_PROMPT
	.if ( eax != S_OK )
		invoke printf, CStr(<"IDebugClient:SetOutputMask failed [%X]",10>), eax
	.endif
endif

;--- set symbol path (in case the image has symbol info)

	invoke GetCurrentDirectory, 260, addr szCurrDir
	invoke vf( pDebugSymbols, IDebugSymbols, SetSymbolPath ), addr szCurrDir
	.if ( eax != S_OK )
		invoke printf, CStr(<"IDebugSymbols:SetSymbolPath failed [%X]",10>), eax
	.endif

;--- try to open the file

	.if ( bDump )
		invoke vf( pDebugClient, IDebugClient, OpenDumpFile ), pszFile
		.if ( eax != S_OK )
			invoke printf, CStr(<"IDebugClient:OpenDumpFile(%s) failed [%X]",10>), pszFile, eax
			jmp exit_
		.endif
	.else
		xor ecx, ecx
		invoke vf( pDebugClient, IDebugClient5, CreateProcess2 ), ecx::ecx, pszFile, addr opbuf, \
			sizeof DEBUG_CREATE_PROCESS_OPTIONS, NULL, NULL
		.if ( eax != S_OK )
			invoke printf, CStr(<"IDebugClient:CreateProcess2(%s) failed [%X]",10>), pszFile, eax
			jmp exit_
		.endif
	.endif
	invoke vf( pDebugControl, IDebugControl, WaitForEvent ), 0, INFINITE

;--- main loop: read input and let the debug engine interpret it

	.while (1)
		invoke vf( pDebugControl, IDebugControl, GetExecutionStatus ), addr dwRead
		.if ( eax != S_OK )
			invoke printf, CStr(<"IDebugControl:GetExecutionStatus failed [%X]",10>), eax
			.break
		.endif

		mov eax, dwRead
;--- exit loop if user pressed 'q'
		.break .if ( eax == DEBUG_STATUS_NO_DEBUGGEE )

;--- if a debuggee runs, call WaitEvent and don't prompt
		.if ( eax == DEBUG_STATUS_GO || eax == DEBUG_STATUS_STEP_OVER || \
			eax == DEBUG_STATUS_STEP_INTO || eax == DEBUG_STATUS_STEP_BRANCH )
			invoke vf( pDebugControl, IDebugControl, WaitForEvent ), 0, INFINITE
			.continue .if ( eax == S_OK )
			invoke printf, CStr(<"IDebugControl:WaitForEvent failed [%X]",10>), eax
		.endif

;--- if debuggee state has changed, display registers and current instruction
		.if ( bStateChanged )
			mov bStateChanged, 0
			invoke vf( pDebugControl, IDebugControl, OutputCurrentState ), DEBUG_OUTCTL_THIS_CLIENT, DEBUG_CURRENT_DEFAULT
		.endif
		
;--- prompt for user input and run it
		invoke vf( pDebugControl, IDebugControl, OutputPrompt ), DEBUG_OUTCTL_THIS_CLIENT, NULL
		invoke vf( pDebugControl, IDebugControl, Input ), addr buffer, sizeof buffer, addr dwRead
		.if ( eax != S_OK )
			invoke printf, CStr(<"IDebugControl:Input failed [%X]",10>), eax
		.elseif ( dwRead )
			invoke vf( pDebugControl, IDebugControl, Execute ), DEBUG_OUTCTL_THIS_CLIENT, addr buffer, 0
		.endif
	.endw

exit_:
	.if ( pDebugClient )
		.if ( pDebugSymbols )
			invoke vf( pDebugSymbols, IUnknown, Release )
		.endif
		.if ( pDebugControl )
			invoke vf( pDebugControl, IUnknown, Release )
		.endif
		invoke vf( pDebugClient, IUnknown, Release )
	.endif
	ret
	align 4

main endp

;--- IDebugEventCallbacks methods

QueryInterface@DebugEventCallbacks proc uses esi edi this_:ptr, pIID:ptr IID, pObject:ptr ptr
	dprintf <"QueryInterface@DebugEventCallbacks called",10>
	mov ecx,4
	mov esi, pIID
	mov edi, offset IID_IDebugEventCallbacks
	repz cmpsd
	jz ok
	mov ecx,4
	mov esi, pIID
	mov edi, offset IID_IUnknown
	repz cmpsd
	jz ok
	mov eax, E_NOINTERFACE
	ret
ok:
	invoke vf( this_, IUnknown, AddRef )
	mov edx, this_
	mov ecx, pObject
	mov [ecx], edx
	mov eax, S_OK
	ret
	align 4
QueryInterface@DebugEventCallbacks endp

AddRef@DebugEventCallbacks proc this_:ptr
	dprintf <"AddRef@DebugEventCallbacks called",10>
	mov ecx, this_
	inc [ecx].CDebugEventCallbacks.cnt
	mov eax, [ecx].CDebugEventCallbacks.cnt
	ret
	align 4
AddRef@DebugEventCallbacks endp

Release@DebugEventCallbacks proc this_:ptr
	dprintf <"Release@DebugEventCallbacks called",10>
	mov ecx, this_
	dec [ecx].CDebugEventCallbacks.cnt
	mov eax, [ecx].CDebugEventCallbacks.cnt
	.if ( eax == 0 )
		invoke free, ecx
		xor eax, eax
	.endif
	ret
	align 4
Release@DebugEventCallbacks endp

GetInterestMask@DebugEventCallbacks proc this_:ptr, Mask_:ptr DWORD
	dprintf <"GetInterestMask@DebugEventCallbacks called",10>
	mov ecx, this_
	mov edx, Mask_
	mov dword ptr [edx], DEBUG_EVENT_BREAKPOINT or DEBUG_EVENT_EXCEPTION or \
		DEBUG_EVENT_CREATE_PROCESS or DEBUG_EVENT_EXIT_PROCESS or \ 
		DEBUG_EVENT_LOAD_MODULE or DEBUG_EVENT_UNLOAD_MODULE or \ 
		DEBUG_EVENT_SYSTEM_ERROR or DEBUG_EVENT_SESSION_STATUS or \
		DEBUG_EVENT_CHANGE_DEBUGGEE_STATE or DEBUG_EVENT_CHANGE_ENGINE_STATE
	mov eax, S_OK
	ret
	align 4
GetInterestMask@DebugEventCallbacks endp

Breakpoint@DebugEventCallbacks proc this_:ptr, brp:ptr ptr
	dprintf <"Breakpoint@DebugEventCallbacks called",10>
	mov eax, DEBUG_STATUS_BREAK
	ret
	align 4
Breakpoint@DebugEventCallbacks endp

Exception@DebugEventCallbacks proc this_:ptr, exc:ptr ptr, firstchance:DWORD
	dprintf <"Exception@DebugEventCallbacks(%X, %u)",10>, exc, firstchance
	mov eax, DEBUG_STATUS_BREAK
	ret
	align 4
Exception@DebugEventCallbacks endp

CreateThread@DebugEventCallbacks proc this_:ptr, handle:qword, dataofs:qword, startofs:qword
	dprintf <"CreateThread@DebugEventCallbacks(%I64X, %I64X, %I64X)",10>, handle, dataofs, startofs
	mov eax, DEBUG_STATUS_NO_CHANGE
	ret
	align 4
CreateThread@DebugEventCallbacks endp

ExitThread@DebugEventCallbacks proc this_:ptr, exitcode:dword
	dprintf <"ExitThread@DebugEventCallbacks(%u)",10>, exitcode
	mov eax, DEBUG_STATUS_NO_CHANGE
	ret
	align 4
ExitThread@DebugEventCallbacks endp

CreateProcess@DebugEventCallbacks proc this_:ptr, filehandle:qword, handle:qword,
		base:qword, modsize:dword, modname:ptr BYTE, imagename:ptr BYTE,
		checksum:dword, timedate:dword, initialthread:qword, threaddata:qword, 
		startofs:qword
	dprintf <"CreateProcess@DebugEventCallbacks(%I64X, %s, %I64X)",10>, base, imagename, startofs
	invoke printf, CStr(<"process entry address: %I64X",10>), startofs
	mov eax, DEBUG_STATUS_BREAK
	ret
	align 4
CreateProcess@DebugEventCallbacks endp

ExitProcess@DebugEventCallbacks proc this_:ptr, exitcode:dword
	dprintf <"ExitProcess@DebugEventCallbacks(%u)",10>, exitcode
	mov eax, DEBUG_STATUS_BREAK
	ret
	align 4
ExitProcess@DebugEventCallbacks endp

LoadModule@DebugEventCallbacks proc this_:ptr, fh:qword, base:qword, modsize:dword, 
		modname:ptr BYTE, imagename:ptr BYTE, checksum:dword, timedate:dword
	dprintf <"LoadModule@DebugEventCallbacks(%I64X, %s)",10>, base, imagename
	mov eax, DEBUG_STATUS_NO_CHANGE
	ret
	align 4
LoadModule@DebugEventCallbacks endp

UnloadModule@DebugEventCallbacks proc this_:ptr, imagename:ptr BYTE, base:qword
	dprintf <"UnloadModule@DebugEventCallbacks(%s, %X)",10>, imagename, base
	mov eax, DEBUG_STATUS_NO_CHANGE
	ret
	align 4
UnloadModule@DebugEventCallbacks endp

SystemError@DebugEventCallbacks proc this_:ptr, error:dword, level:dword
	dprintf <"SystemError@DebugEventCallbacks called",10>
	mov eax, DEBUG_STATUS_NO_CHANGE
	ret
	align 4
SystemError@DebugEventCallbacks endp

SessionStatus@DebugEventCallbacks proc this_:ptr, status:dword
	dprintf <"SessionStatus@DebugEventCallbacks(%X)",10>, status
	mov eax, DEBUG_STATUS_NO_CHANGE	;return value is ignored
	ret
	align 4
SessionStatus@DebugEventCallbacks endp

ChangeDebuggeeState@DebugEventCallbacks proc this_:ptr, flags:dword, arg:qword
	dprintf <"ChangeDebuggeeState@DebugEventCallbacks(%X, %I64X)",10>, flags, arg
	mov bStateChanged, 1
	mov eax, DEBUG_STATUS_NO_CHANGE	;return value is ignored
	ret
	align 4
ChangeDebuggeeState@DebugEventCallbacks endp

ChangeEngineState@DebugEventCallbacks proc this_:ptr, flags:dword, arg:qword
	dprintf <"ChangeEngineState@DebugEventCallbacks(%X, %I64X)",10>, flags, arg
	mov eax, DEBUG_STATUS_NO_CHANGE	;return value is ignored
	ret
	align 4
ChangeEngineState@DebugEventCallbacks endp

ChangeSymbolState@DebugEventCallbacks proc this_:ptr, flags:dword, arg:qword
	dprintf <"ChangeSymbolState@DebugEventCallbacks(%X, %I64X)",10>, flags, arg
	mov eax, DEBUG_STATUS_NO_CHANGE	;return value is ignored
	ret
	align 4
ChangeSymbolState@DebugEventCallbacks endp

;--- IDebugInputCallbacks methods

QueryInterface@DebugInputCallbacks proc uses esi edi this_:ptr, pIID:ptr IID, pObject:ptr ptr
	dprintf <"QueryInterface@DebugInputCallbacks called",10>
	mov ecx,4
	mov esi, pIID
	mov edi, offset IID_IDebugInputCallbacks
	repz cmpsd
	jz ok
	mov ecx,4
	mov esi, pIID
	mov edi, offset IID_IUnknown
	repz cmpsd
	jz ok
	mov eax, E_NOINTERFACE
	ret
ok:
	invoke vf( this_, IUnknown, AddRef )
	mov edx, this_
	mov ecx, pObject
	mov [ecx], edx
	mov eax, S_OK
	ret
	align 4
QueryInterface@DebugInputCallbacks endp

AddRef@DebugInputCallbacks proc this_:ptr
	dprintf <"AddRef@DebugInputCallbacks called",10>
	mov ecx, this_
	inc [ecx].CDebugInputCallbacks.cnt
	mov eax, [ecx].CDebugInputCallbacks.cnt
	ret
	align 4
AddRef@DebugInputCallbacks endp

Release@DebugInputCallbacks proc this_:ptr
	dprintf <"Release@DebugInputCallbacks called",10>
	mov ecx, this_
	dec [ecx].CDebugInputCallbacks.cnt
	mov eax, [ecx].CDebugInputCallbacks.cnt
	.if ( eax == 0 )
		invoke free, ecx
		xor eax, eax
	.endif
	ret
	align 4
Release@DebugInputCallbacks endp

StartInput@DebugInputCallbacks proc uses ebx this_:ptr, BufferSize:DWORD

local dwRead:dword
local dwEsp:dword

;	dprintf <"StartInput@DebugInputCallbacks(%u)",10>, BufferSize
	invoke GetStdHandle, STD_INPUT_HANDLE
	mov ebx, eax
	mov dwEsp, esp
	mov eax, BufferSize
	add eax, 3
	and al, 0FCh
	sub esp, eax
	mov edx, esp
	invoke ReadFile, ebx, edx, BufferSize, addr dwRead, NULL
	.if ( eax && dwRead )
		mov ecx, dwRead
		.if ( byte ptr [esp+ecx-1] == 10 )
			mov byte ptr [esp+ecx-1],0
		.endif
		invoke vf( pDebugControl, IDebugControl, ReturnInput ), esp
		.if ( eax != S_OK )
			invoke printf, CStr(<"IDebugControl:ReturnInput failed [%X]",10>), eax
		.endif
	.endif
	mov esp, dwEsp
	mov eax, S_OK
	ret
	align 4
StartInput@DebugInputCallbacks endp

EndInput@DebugInputCallbacks proc this_:ptr
;	dprintf <"EndInput@DebugInputCallbacks called",10>
	mov eax, S_OK
	ret
	align 4
EndInput@DebugInputCallbacks endp

;--- IDebugOutputCallbacks methods

QueryInterface@DebugOutputCallbacks proc uses esi edi this_:ptr, pIID:ptr IID, pObject:ptr ptr
	dprintf <"QueryInterface@DebugOutputCallbacks called",10>
	mov ecx,4
	mov esi, pIID
	mov edi, offset IID_IDebugOutputCallbacks
	repz cmpsd
	jz ok
	mov ecx,4
	mov esi, pIID
	mov edi, offset IID_IUnknown
	repz cmpsd
	jz ok
	mov eax, E_NOINTERFACE
	ret
ok:
	invoke vf( this_, IUnknown, AddRef )
	mov edx, this_
	mov ecx, pObject
	mov [ecx], edx
	mov eax, S_OK
	ret
	align 4
QueryInterface@DebugOutputCallbacks endp

AddRef@DebugOutputCallbacks proc this_:ptr
	dprintf <"AddRef@DebugOutputCallbacks called",10>
	mov ecx, this_
	inc [ecx].CDebugOutputCallbacks.cnt
	mov eax, [ecx].CDebugOutputCallbacks.cnt
	ret
	align 4
AddRef@DebugOutputCallbacks endp

Release@DebugOutputCallbacks proc this_:ptr
	dprintf <"Release@DebugOutputCallbacks called",10>
	mov ecx, this_
	dec [ecx].CDebugOutputCallbacks.cnt
	mov eax, [ecx].CDebugOutputCallbacks.cnt
	.if ( eax == 0 )
		invoke free, ecx
		xor eax, eax
	.endif
	ret
	align 4
Release@DebugOutputCallbacks endp

Output@DebugOutputCallbacks proc c this_:ptr, Mask_:DWORD, Text:ptr BYTE
;	dprintf <"Output@DebugOutputCallbacks called",10>
	invoke printf, CStr(<"%s">), Text
	mov eax, S_OK
	ret
	align 4
Output@DebugOutputCallbacks endp

__getmainargs proto c :ptr DWORD, :ptr DWORD, :ptr DWORD, :DWORD, :DWORD

start proc c

local argc:dword
local argv:dword
local envp:dword
local _si:STARTUPINFO

	invoke __getmainargs, addr argc, addr argv, addr envp, 0, addr _si
	invoke main, argc, argv
	invoke ExitProcess,eax

start endp

	end start
